#include <ipxe/settings.h>
#include <ipxe/tables.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <ipxe/timer.h>
#include <ipxe/netdevice.h>
#include <ipxe/driver_settings.h>
#include <ipxe/in.h>
#include <ipxe/infiniband.h>
#include <ipxe/vlan.h>
#include <ipxe/iscsi.h>
#include "hermon.h"
#include "hermon_settings.h"

/* Hermon setting operation */
struct hermon_setting_operation {
	/** Setting */
	const struct setting *setting;
	int ( *applies ) ( struct settings *settings );
	/** Store setting (or NULL if not supported)
	 *
	 * @v netdev		Network device
	 * @v data		Setting data, or NULL to clear setting
	 * @v len		Length of setting data
	 * @ret rc		Return status code
	 */
	int ( * store ) ( struct settings *settings, const void *data,
			  size_t len, const struct setting *setting );
};

struct hermon_device_name {
	/** Device ID */
	uint16_t device_id;
	/** Device name */
	char* name;
};

struct net_device* get_netdev_from_settings ( struct settings *settings ) {
	struct settings *main_settings = settings;
        struct generic_settings *generic;
        struct driver_settings *driver_set;
        struct hermon_port *port;
	struct net_device *trunk;
	struct net_device *netdev;

	while ( main_settings->default_scope != &port_scope )
		main_settings = main_settings->parent;

	generic = container_of ( main_settings, struct generic_settings, settings);
	driver_set = container_of ( generic, struct driver_settings, generic_settings);
	port = container_of ( driver_set, struct hermon_port, driver_settings);
	trunk = port->netdev;

	netdev =  vlan_present( trunk );
	if ( netdev )
		return netdev;

	return trunk;

}

int netdevice_settings_store ( struct settings *settings, const struct setting *setting,
				const void *data, size_t len ) {
	struct settings *port_settings = settings;
        struct net_device *netdev = get_netdev_from_settings ( settings );
	char dhcp_ip_en[10] = { 0 };
        int rc;

	/* Check if boot protocol is iSCSI, if not then don't store system settings */
	if ( ! is_protocol_iscsi ( settings ) )
		return SUCCESS;

	while ( port_settings->default_scope != &port_scope )
		port_settings = port_settings->parent;

	fetchf_setting ( port_settings, &dhcp_ip_setting, NULL, NULL, dhcp_ip_en, sizeof ( dhcp_ip_en ) );
	if ( dhcp_ip_en[0] == 'E' )
		return SUCCESS;

	if ( ( rc = store_setting ( netdev_settings ( netdev ), setting, data, len ) ) == 0 )
		return SUCCESS;

        return INVALID_INPUT;
}

int ipv4_address_store ( struct settings *settings , const void *data,
		    size_t len, const struct setting *setting __unused)  {
	return netdevice_settings_store ( settings, &ip_setting, data, len );
}

int subnet_mask_store ( struct settings *settings , const void *data,
                    size_t len, const struct setting *setting __unused)  {
        return netdevice_settings_store( settings, &netmask_setting, data, len );
}

int ipv4_gateway_store ( struct settings *settings , const void *data,
                    size_t len, const struct setting *setting __unused)  {
        return netdevice_settings_store( settings, &gateway_setting, data, len );
}

int ipv4_dns_store ( struct settings *settings , const void *data,
                    size_t len, const struct setting *setting __unused)  {
        return netdevice_settings_store( settings, &dns_setting, data, len );
}

int iscsi_settings_store ( struct settings *settings, const struct setting *setting,
				const void *data, size_t len ) {
	struct settings *port_settings = settings;
	struct net_device *netdev = get_netdev_from_settings ( settings );
	char dhcp_params[10] = { 0 };
	int rc;

	while ( port_settings->default_scope != &port_scope )
		port_settings = port_settings->parent;

	fetchf_setting ( port_settings, &dhcp_iscsi_setting, NULL, NULL, dhcp_params, sizeof ( dhcp_params ) );
	if ( dhcp_params[0] == 'E' )
		return SUCCESS;

	if ( ( rc = store_setting ( netdev_settings ( netdev ), setting, data, len ) ) )
		return INVALID_INPUT;

	return SUCCESS;
}

int iscsi_init_settings_check_and_store ( struct settings *settings, const struct setting *setting,
				const void *data, size_t len, unsigned int max_len, unsigned int min_len ) {
	if ( len != 0 &&  ( len > max_len || len < min_len ) )
		return INVALID_INPUT;

	return iscsi_settings_store ( settings, setting, data, len );
}

int iscsi_tgt_settings_check_and_store ( struct settings *settings, const struct setting *setting,
				const void *data, size_t len, unsigned int max_len, unsigned int min_len ) {
	struct settings *port_settings = settings;
	char connect[10] = { 0 };
	char dhcp_params[10] = { 0 };

	if ( len != 0 && ( len > max_len || len < min_len ) )
		return INVALID_INPUT;

	while ( port_settings->default_scope != &port_scope )
		port_settings = port_settings->parent;

	/* First check if connect is enabled, if not then don't save in system settings */
	fetchf_setting ( port_settings, &connect_setting, NULL, NULL, connect, sizeof ( connect ) );
	fetchf_setting ( settings, &dhcp_iscsi_setting, NULL, NULL, dhcp_params, sizeof ( dhcp_params ) );
	if ( ( connect[0] != 'E' ) || ( dhcp_params[0] == 'E' ) )
		return SUCCESS;

	return iscsi_settings_store ( settings, setting, data, len );
}

int iscsi_initiator_name_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting __unused ) {
	return iscsi_init_settings_check_and_store ( settings, &initiator_iqn_setting,
			data, len, MAX_ISCSI_NAME, MIN_ISCSI_NAME );
}

int iscsi_initiator_chapid_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting __unused ) {
        return iscsi_init_settings_check_and_store ( settings, &reverse_username_setting,
			data, len, MAX_CHAP_ID , MIN_CHAP_ID);
}

int iscsi_initiator_chapsec_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting __unused ) {
        return iscsi_init_settings_check_and_store ( settings, &reverse_password_setting,
			data, len, MAX_CHAP_SECRET , MIN_CHAP_SECRET);
}

int iscsi_target_chapid_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting __unused ) {
        return iscsi_tgt_settings_check_and_store ( settings, &username_setting,
			data, len, MAX_CHAP_ID , MIN_CHAP_ID);
}

int iscsi_target_chapsec_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting __unused ) {
        return iscsi_tgt_settings_check_and_store ( settings, &password_setting,
			data, len, MAX_CHAP_SECRET , MIN_CHAP_SECRET);
}

int iscsi_tgt_store_root_path ( struct settings *settings, const struct setting *setting,
			const void *data, size_t len ) {
	struct net_device *netdev;
	char old[256];
	int fetch_rc, rc;

	if ( ! settings )
		return INVALID_INPUT;

	/* Save previous value in case of error */
	fetch_rc = fetchf_setting ( settings, setting, NULL, NULL, old, sizeof ( old ) );
	if ( ( rc = generic_settings_store ( settings, setting, data, len ) ) )
		return rc;

	netdev = get_netdev_from_settings ( settings );
	if ( ( rc = root_path_store ( netdev_settings ( netdev ), settings ) ) ) {
		if ( fetch_rc > 0 )
			generic_settings_store ( settings, setting, old, sizeof ( old ) );
		return rc;
	}

	return 0;
}

int target_ip_store ( struct settings *settings, const void *data,
                    size_t len, const struct setting *setting ) {
	return iscsi_tgt_store_root_path ( settings, setting, data, len );
}

int target_iscsi_name_store ( struct settings *settings , const void *data,
                    size_t len , const struct setting *setting ) {
	if ( data && len > MAX_ISCSI_NAME ) {
		return INVALID_INPUT;
	}

	return iscsi_tgt_store_root_path ( settings, setting, data, len );
}

int target_tcp_port_store ( struct settings *settings , const void *data,
                    size_t len, const struct setting *setting ) {
	int32_t input = 0;

	if ( data ) {
		input = ntohl ( *( ( int32_t * ) data ) );

		if ( input > MAX_TCP_PORT || input < MIN_TCP_PORT )
			return INVALID_INPUT;
	}

	return iscsi_tgt_store_root_path ( settings, setting, data, len );
}

int target_boot_lun_store ( struct settings *settings , const void *data,
                    size_t len, const struct setting *setting ) {
	const char *buffer = ( const char * ) data;
	char *max = MAX_BOOT_LUN_STR;
	unsigned int i;

	if ( len > strlen ( MAX_BOOT_LUN_STR ) )
		return INVALID_INPUT;

	for ( i = 0; i < len; i++ ) {
		/* Check that all the characters are digits */
		if ( ( buffer[i] > '9' ) || ( buffer[i] < '0' ) )
			return INVALID_INPUT;
		/* Check that it doesn't exceed the maximum */
		if ( ( len == strlen( MAX_BOOT_LUN_STR ) ) && ( buffer[i] > max[i] ) )
			return INVALID_INPUT;
	}

	return iscsi_tgt_store_root_path ( settings, setting, data, len );
}


struct driver_settings* get_driver_settings_from_settings ( struct settings *settings ) {
	struct settings *main_settings = settings;
	struct generic_settings *generic;
	struct driver_settings *driver_settings;

	/* If these are not hermon_settings, needs to get to main settings */
	if ( ( settings->default_scope != &main_scope ) && ( settings->default_scope != &fw_scope ) ) {
		while ( main_settings->default_scope != &port_scope )
			main_settings = main_settings->parent;
	} else {
		if ( settings->default_scope == &fw_scope )
			main_settings = main_settings->parent;
	}

	generic = container_of ( main_settings, struct generic_settings, settings);
	driver_settings = container_of ( generic, struct driver_settings, generic_settings);

	return driver_settings;
}

struct hermon_port* get_hermon_port_from_settings ( struct settings *settings ) {
	struct settings *port_settings = settings;

	if ( ! settings )
		return NULL;

	while ( port_settings->default_scope != &port_scope )
		port_settings = port_settings->parent;

	struct driver_settings *driver_settings = get_driver_settings_from_settings ( port_settings );
	struct hermon_port* port = container_of ( driver_settings, struct hermon_port, driver_settings );
	return port;
}

int hermon_wol_applies ( struct settings *settings ) {
	struct hermon_port *port = get_hermon_port_from_settings ( settings );
	if ( ! port )
		return DOES_NOT_APPLY;
	struct hermon *hermon = ib_get_drvdata ( port->ibdev );

	if ( ( port->ibdev->port == 1 ) && hermon->cap.nv_config_flags.nv_config_wol_port1 )
		return APPLIES;

	if ( ( port->ibdev->port == 2 ) && hermon->cap.nv_config_flags.nv_config_wol_port2 )
		return APPLIES;

	return DOES_NOT_APPLY;
}

int hermon_blink_leds_applies ( struct settings *settings ) {
	struct hermon_port *port = get_hermon_port_from_settings ( settings );
	if ( ! port )
		return DOES_NOT_APPLY;
        struct hermon *hermon = ib_get_drvdata ( port->ibdev );

	return hermon->cap.port_beacon;
}

int hermon_vm_applies (struct settings *settings ) {
	struct driver_settings* driver_settings = get_driver_settings_from_settings ( settings );
	struct hermon *hermon = container_of ( driver_settings, struct hermon, hermon_settings );

	return hermon->cap.nv_config_flags.nv_config_sriov_en;
}

static int hermon_nv_write_blink_leds ( struct hermon *hermon,
					struct hermon_port *port, int duration ) {
	struct hermon_nvmem_port_conf *conf = & ( port->port_nv_conf );
	int rc;

	conf->blink_leds = duration;

	if ( ( rc = hermon_flash_write_blink_leds ( hermon, port ) ) != 0 ) {
		printf ( "Failed to write general settings to the flash (rc = %d)\n", rc );
		return rc;
	}

	return 0;
}

int blink_leds_store ( struct settings *settings, const void *data,
			  size_t len,const  struct setting *setting ) {
	struct hermon_port *port;
	struct hermon *hermon;
	int32_t input;
	int rc;

	if ( !data  || ntohl ( * ( int32_t * ) data ) == 0 )
		return SUCCESS;

	input = ntohl ( * ( int32_t * ) data );

	if ( input > MAX_BLINK_LEDS || input < 0 )
		return INVALID_INPUT;

	port = get_hermon_port_from_settings ( settings );
	if ( !port )
		return INVALID_INPUT;
	hermon = ib_get_drvdata ( port->ibdev );
	generic_settings_store ( settings, setting, data, len );

	if ( ( rc = hermon_open ( hermon ) ) != 0 ) {
		printf ( " Couldn't open hermon (rc = %d) \n", rc );
		return INVALID_INPUT; /* TODO - change to einval */
	}

	if ( (rc =  hermon_nv_write_blink_leds ( hermon, port, input ) ) != 0 ) {
		DBGC ( hermon, "Failed to store blink leds (rc = %d)\n",        rc );
	}

	/* Close hardware */
	hermon_close ( hermon );

	return STORED;
}

#define MAX_SETTING_STR 128
int virt_mode_restore ( struct settings *settings, const void *data,
						size_t len __unused, const struct setting *setting ) {
	struct driver_settings* driver_settings = get_driver_settings_from_settings ( settings );
        struct hermon *hermon = container_of ( driver_settings, struct hermon, hermon_settings );

	/* If setting is deleted, set default */
	if ( !data ) {
		if ( hermon->defaults.sriov_en )
			generic_settings_store ( settings, setting, STR_SRIOV, strlen ( STR_SRIOV ) + 1 );
		else
			generic_settings_store ( settings, setting, STR_NONE, strlen ( STR_NONE ) + 1 );
		return STORED;
	}

	return SUCCESS;
}

#define RESTORE_ENABLED_DISABLED( defaultdata )		do {								\
	struct hermon_port *port;											\
		if ( !data ) {												\
			port = get_hermon_port_from_settings ( settings );						\
			if ( ! port )																							\
				return SUCCESS;																			\
			if ( port->defaults.defaultdata )								\
				generic_settings_store ( settings, setting, STR_ENABLE, strlen ( STR_ENABLE ) + 1 );	\
			else												\
				generic_settings_store ( settings, setting, STR_DISABLE, strlen ( STR_DISABLE ) + 1 );	\
			return STORED;											\
		}													\
		return SUCCESS;												\
	} while ( 0 )


int virt_num_check_or_restore ( struct settings *settings, const void *data,
				size_t len __unused, const struct setting *setting ) {
	struct driver_settings* driver_settings = get_driver_settings_from_settings ( settings );
	struct hermon *hermon = container_of ( driver_settings, struct hermon, hermon_settings );
	int32_t maximum, number;
	int rc;

	/* If setting is deleted, set default */
	if ( !data ) {
		number = ntohl ( hermon->defaults.total_vfs  );
		generic_settings_store ( settings, setting, &number, sizeof ( number ) );
		return STORED;
	}

	number = ntohl ( * ( int32_t * ) data );
	if ( number < 0 )
		return INVALID_INPUT;

	if ( ( rc = generic_settings_fetch ( settings, &virt_num_max_setting,
						&maximum, sizeof ( maximum ) ) ) <= 0 )
		return rc;

	maximum = htonl ( maximum );
	if ( number >= maximum )
		return INVALID_INPUT;

	return SUCCESS;
}

int boot_protocol_restore ( struct settings *settings, const void *data,
			size_t len __unused, const struct setting *setting ) {
	struct hermon_port *port;

        /* If setting is deleted, set default */
        if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;

		/** Assuming boot protocol is 0-none, 1-pxe, 2-iscsi */
		switch ( port->defaults.boot_protocol ) {
			case 2  : generic_settings_store ( settings, setting, STR_ISCSI, strlen ( STR_ISCSI ) + 1 );
				   break;
			case 1 : generic_settings_store ( settings, setting, STR_PXE, strlen ( STR_PXE ) + 1 );
				   break;
			default : generic_settings_store ( settings, setting, STR_NONE, strlen ( STR_NONE ) + 1 );
				   break;
		}
		return STORED;
	}

	return SUCCESS;
}

int virt_lan_restore ( struct settings *settings, const void *data,
			size_t len __unused, const struct setting *setting ) {
	RESTORE_ENABLED_DISABLED ( boot_vlan_en );
}

int virt_id_check_or_restore ( struct settings *settings, const void *data,
			size_t len __unused ,const  struct setting *setting ) {
	struct hermon_port *port;
	int32_t number;

        /* If setting is deleted, set default */
	if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;
		number = ntohl ( port->defaults.boot_vlan  );
		generic_settings_store ( settings, setting, &number, sizeof ( number ) );
		return STORED;
	}

	number = ntohl ( * ( int32_t * ) data );

	if ( number > MAX_VIRTUAL_ID || number < MIN_VIRTUAL_ID )
		return INVALID_INPUT;

	return SUCCESS;
}

int opt_rom_restore ( struct settings *settings, const void *data,
                    size_t len __unused ,const  struct setting *setting ) {
	RESTORE_ENABLED_DISABLED ( boot_option_rom_en );
}

int boot_retries_restore ( struct settings *settings, const void *data,
                    size_t len __unused ,const  struct setting *setting ) {
	struct hermon_port *port;

	const char *retries;
	/* If setting is deleted, set default */
	if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;
		retries = driver_settings_get_boot_ret_str ( port->defaults.boot_retry_count );
		generic_settings_store ( settings, setting, retries, 10 );
		return STORED;
	}

	return SUCCESS;
}

int wol_restore ( struct settings *settings, const void *data,
                    size_t len __unused ,const  struct setting *setting ) {
        RESTORE_ENABLED_DISABLED ( en_wol_magic );
}

int dhcp_ip_restore ( struct settings *settings, const void *data,
		size_t len __unused ,const  struct setting *setting ) {
	struct hermon_port *port;
	if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;
		if ( port->defaults.iscsi_ipv4_dhcp_en == 1 )
			generic_settings_store ( settings, setting, STR_ENABLE, strlen ( STR_ENABLE ) + 1 );
		else
			generic_settings_store ( settings, setting, STR_DISABLE, strlen ( STR_DISABLE ) + 1 );
		return STORED;
	}
	return SUCCESS;
}

int dhcp_iscsi_restore ( struct settings *settings, const void *data,
		size_t len __unused ,const  struct setting *setting ) {
	struct hermon_port *port;
	if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;
		if ( port->defaults.iscsi_dhcp_params_en == 1 )
			generic_settings_store ( settings, setting, STR_ENABLE, strlen ( STR_ENABLE ) + 1 );
		else
			generic_settings_store ( settings, setting, STR_DISABLE, strlen ( STR_DISABLE ) + 1 );
		return STORED;
	}
	return SUCCESS;
}

int connect_restore ( struct settings *settings, const void *data,
                    size_t len __unused ,const  struct setting *setting ) {
	struct hermon_port *port;
	if ( !data ) {
		port = get_hermon_port_from_settings ( settings );
		if ( ! port )
			return SUCCESS;
		if ( ( port->port_nv_conf.nic.boot_conf.legacy_boot_prot == BOOT_PROTOCOL_ISCSI ) &&
		     ( port->defaults.iscsi_dhcp_params_en == 0 ) )
			generic_settings_store ( settings, setting, STR_ENABLE, strlen ( STR_ENABLE ) + 1 );
		else
			generic_settings_store ( settings, setting, STR_DISABLE, strlen ( STR_DISABLE ) + 1 );
		return STORED;
	}
	return SUCCESS;
}

static struct hermon_setting_operation hermon_setting_operations[] = {
	{ &virt_mode_setting,           &hermon_vm_applies ,            &virt_mode_restore },
	{ &virt_num_setting,		NULL, 				&virt_num_check_or_restore },
	{ &virt_num_max_setting,	NULL, 				NULL },
	{ &blink_leds_setting,		&hermon_blink_leds_applies,	&blink_leds_store },
	{ &device_name_setting, 	NULL,				NULL },
	{ &chip_type_setting,		NULL,				NULL },
	{ &pci_id_setting,		NULL, 				NULL },
	{ &bus_dev_fun_setting,		NULL,				NULL },
	{ &mac_add_setting,		NULL,				NULL },
	{ &virt_mac_setting,		NULL,				NULL },
	{ &flex_version_setting,	NULL,				NULL },
	{ &fw_version_setting,		NULL, 				NULL },
	{ &boot_protocol_setting, 	NULL,				&boot_protocol_restore },
	{ &virt_lan_setting,		NULL,				&virt_lan_restore },
	{ &virt_id_setting,		NULL,				&virt_id_check_or_restore },
	{ &opt_rom_setting,		NULL,				&opt_rom_restore },
	{ &boot_retries_setting,	NULL,				&boot_retries_restore },
	{ &wol_setting,			&hermon_wol_applies,		&wol_restore },
	{ &dhcp_ip_setting,		NULL,				&dhcp_ip_restore },
	{ &dhcp_iscsi_setting,		NULL,				&dhcp_iscsi_restore },
	{ &ip_ver_setting, 		NULL, 				NULL },
	{ &ipv4_add_setting,		NULL,				&ipv4_address_store },
	{ &subnet_mask_setting, 	NULL, 				&subnet_mask_store },
	{ &ipv4_gateway_setting,	NULL, 				&ipv4_gateway_store },
	{ &ipv4_dns_setting,		NULL, 				&ipv4_dns_store },
	{ &iscsi_init_name_setting,	NULL, 				&iscsi_initiator_name_store },
	{ &init_chapid_setting,		NULL,				&iscsi_initiator_chapid_store },
	{ &init_chapsec_setting,	NULL,				&iscsi_initiator_chapsec_store },
	{ &connect_setting,		NULL,				&connect_restore },
	{ &target_ip_setting,		NULL,				&target_ip_store },
	{ &tcp_port_setting,		NULL,				&target_tcp_port_store },
	{ &boot_lun_setting,		NULL,				&target_boot_lun_store },
	{ &iscsi_target_name_setting,	NULL,				&target_iscsi_name_store },
	{ &target_chapid_setting,	NULL,				&iscsi_target_chapid_store },
	{ &target_chapsec_setting,	NULL,				&iscsi_target_chapsec_store },
};


static int is_modified ( struct list_head *list,const char *name ) {
	struct extended_setting *ext;

	list_for_each_entry ( ext, list, modified ) {
		if ( strcmp (ext->setting->name, name ) == 0 )
			return 1;
	}
	return 0;
}

int hermon_store ( struct settings *settings,
		   const struct setting *setting,
		   const void *data, size_t len ) {

	struct hermon_setting_operation *op;
	struct extended_setting *ext;
	struct extended_setting *new_ext;
	struct driver_settings *driver_settings;
	unsigned int i;
	int rc = 0;

	/* Handle hermon specific settings */
	for ( i = 0 ; i < ( sizeof ( hermon_setting_operations ) /
			    sizeof ( hermon_setting_operations[0] ) ) ; i++ ) {
		op = &hermon_setting_operations[i];
		if ( setting_cmp ( setting, op->setting ) == 0 ) {
			if ( op->store ) {
				if ( ( rc = op->store ( settings, data, len, setting ) ) == INVALID_INPUT )
					return rc;
			}
			break;
		}
	}

	if ( rc != STORED )
		rc = generic_settings_store ( settings, setting, data, len );
	else
		rc = SUCCESS;

	/* Save setting in modified list - only for driver settings */
	driver_settings = get_driver_settings_from_settings ( settings );
	if ( rc == SUCCESS && driver_settings->store_menu ) {
		if ( ! is_modified ( &driver_settings->modified_list, setting->name ) ) {
			new_ext = malloc ( sizeof ( struct extended_setting ) );
			ext = find_extended ( setting );
			memcpy ( new_ext, ext, sizeof ( struct extended_setting ) );
			list_add ( &new_ext->modified, &driver_settings->modified_list );
		}
	}
	return rc;
}

int hermon_applies ( struct settings *settings, const struct setting *setting) {
	struct hermon_setting_operation *op;
	unsigned int i, rc;

	for ( i = 0 ; i < ( sizeof ( hermon_setting_operations ) /
			    sizeof ( hermon_setting_operations[0] ) ) ; i++ ) {
		op = &hermon_setting_operations[i];
		if ( setting_cmp ( setting, op->setting ) == 0 ) {
			if ( setting->scope == settings->default_scope) {
				if ( op->applies ) {
					if ( ( rc = op->applies ( settings ) ) )
						return APPLIES;
					else
						return DOES_NOT_APPLY;
				} else {
					return APPLIES;
				}
				break;
			}
		}
	}
	return DOES_NOT_APPLY;
}

struct settings_operations hermon_settings_operations = {
	.store = &hermon_store,
	.fetch = &generic_settings_fetch,
	.applies = &hermon_applies,
};

#define HERMON_FETCH_SETTING( settings, setting )					\
	do {										\
		memset ( buf, 0, sizeof ( buf ) );					\
		if ( ( rc = fetchf_setting ( settings, setting , NULL, NULL, buf,	\
					sizeof ( buf ) ) ) <= 0 ) {			\
			DBGC ( hermon, "Failed to fetch %s setting (rc = %d)\n",	\
				( setting )->name, rc );				\
		}									\
	} while ( 0 )

static int hermon_nv_write_nic_config ( struct hermon *hermon, struct hermon_port *port ) {
	struct settings *settings = & ( port->driver_settings.generic_settings.settings );
	struct hermon_nic_conf *nic_conf = & ( port->port_nv_conf.nic );
	union hermon_nic_boot_conf *boot_conf = & ( nic_conf->boot_conf );
	char buf[255] = {0};
	int rc, modified = 0;

	HERMON_FETCH_SETTING ( settings, &boot_protocol_setting );
	if ( rc > 0 ) {
		modified = 1;

		if ( buf[0] == 'N' )
			boot_conf->legacy_boot_prot = BOOT_PROTOCOL_NONE;
		else if ( buf[0] == 'P' )
			boot_conf->legacy_boot_prot = BOOT_PROTOCOL_PXE;
		else
			boot_conf->legacy_boot_prot = BOOT_PROTOCOL_ISCSI;
	}

	HERMON_FETCH_SETTING ( settings, &virt_lan_setting );
	if ( rc > 0 ) {
		modified = 1;
		boot_conf->en_vlan = ( buf[0] == 'E' );
	}

	HERMON_FETCH_SETTING ( settings, &virt_id_setting );
	if ( rc > 0 ) {
		modified = 1;
		boot_conf->vlan_id = strtoul ( buf, NULL, 10 );
	}

	HERMON_FETCH_SETTING ( settings, &opt_rom_setting );
	if ( rc > 0 ) {
		modified = 1;
		boot_conf->en_option_rom = ( buf[0] == 'E' );
	}

	boot_conf->boot_retry_count = driver_settings_get_boot_ret_val ( settings );
	modified = 1; /* set modified 1 for now */

	if ( modified ) {
		if ( ( rc = hermon_flash_write_nic_boot_config ( hermon, port ) ) )
			DBGC ( hermon, "Failed to flash the nic boot config (rc = %d)\n", rc );
	}

	HERMON_FETCH_SETTING ( settings, &wol_setting );
	if ( rc > 0 ) {
		nic_conf->wol_conf.en_wol_magic = ( buf[0] == 'E' );

		if ( ( rc = hermon_flash_write_nic_wol_config ( hermon, port ) ) )
			DBGC ( hermon, "Failed to flash the nic WOL config (rc = %d)\n", rc );
	}

	return 0;
}

#define HERMON_FETCH_ISCSI_AND_UPDATE( settings, setting, field, is_valid, mask )	\
	do {										\
		memset ( buf, 0, sizeof ( buf ) );					\
		if ( ( rc = fetchf_setting ( settings, setting , NULL, NULL, buf,	\
					sizeof ( buf ) ) ) <= 0 ) {			\
			( is_valid ) &= ~ ( mask );					\
			DBGC ( hermon, "Failed to fetch %s setting (rc = %d)\n",	\
					( setting )->name, rc );			\
		}									\
		( is_valid ) |= ( mask );						\
		memcpy ( field, buf, sizeof ( field ) );				\
	} while ( 0 )


static void hermon_nv_write_iscsi_init_dhcp ( struct hermon_port *port ) {
	struct hermon *hermon = ib_get_drvdata ( port->ibdev );
	struct settings *settings = & ( port->driver_settings.generic_settings.settings );
	union hermon_iscsi_init_dhcp_conf *conf = & ( port->port_nv_conf.iscsi.init_dhcp_conf );
	char buf[255] = {0};
	int rc;

	memset ( buf, 0, sizeof ( buf ) );
	if ( ( rc = fetchf_setting ( settings, &dhcp_ip_setting, NULL, NULL, buf,sizeof ( buf ) ) ) <= 0 ) {
		DBGC ( hermon, "Failed to fetch %s setting (rc = %d)\n", dhcp_ip_setting.name, rc );
	}
	conf->ipv4_dhcp_en = ( buf[0] == 'E' );

	memset ( buf, 0, sizeof ( buf ) );
	if ( ( rc = fetchf_setting ( settings, &dhcp_iscsi_setting, NULL, NULL, buf,sizeof ( buf ) ) ) <= 0 ) {
		DBGC ( hermon, "Failed to fetch %s setting (rc = %d)\n", dhcp_iscsi_setting.name, rc );
	}
	conf->dhcp_iscsi_en = ( buf[0] == 'E' );
}

static void hermon_nv_write_iscsi_first_tgt ( struct hermon_port *port ) {
	struct hermon *hermon = ib_get_drvdata ( port->ibdev );
	struct settings *settings = & ( port->driver_settings.generic_settings.settings );
	struct hermon_iscsi_target_params *conf = & ( port->port_nv_conf.iscsi.first_tgt_params );
	char buf[255] = {0};
	int rc;

	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &connect_setting, conf->connect,
			conf->valid_mask, ISCSI_TARGET_CONNECT_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &target_ip_setting, conf->ip_addr,
			conf->valid_mask, ISCSI_TARGET_IP_ADDR_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &tcp_port_setting, conf->tcp_port,
			conf->valid_mask, ISCSI_TARGET_TCP_PORT_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &boot_lun_setting, conf->lun,
			conf->valid_mask, ISCSI_TARGET_LUN_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &iscsi_target_name_setting, conf->name,
			conf->valid_mask, ISCSI_TARGET_NAME_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &target_chapid_setting, conf->chap_id,
			conf->valid_mask, ISCSI_TARGET_CHAP_ID_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &target_chapsec_setting, conf->chap_pass,
			conf->valid_mask, ISCSI_TARGET_CHAP_PASS_MASK );
}

static void hermon_nv_write_iscsi_initiator ( struct hermon_port *port ) {
	struct hermon *hermon = ib_get_drvdata ( port->ibdev );
	struct settings *settings = & ( port->driver_settings.generic_settings.settings );
	struct hermon_iscsi_initiator_params *conf = & ( port->port_nv_conf.iscsi.initiator_params );
	char buf[255] = {0};
	int rc;

	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &ipv4_add_setting, conf->ipv4_addr,
			conf->valid_mask, ISCSI_INITIATOR_IP_ADDR_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &subnet_mask_setting, conf->subnet,
			conf->valid_mask, ISCSI_INITIATOR_SUBNET_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &ipv4_gateway_setting, conf->gateway,
			conf->valid_mask, ISCSI_INITIATOR_GATEWAY_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &ipv4_dns_setting, conf->primary_dns,
			conf->valid_mask, ISCSI_INITIATOR_PRIME_DNS_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &iscsi_init_name_setting, conf->name,
			conf->valid_mask, ISCSI_INITIATOR_NAME_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &init_chapid_setting, conf->chap_id,
			conf->valid_mask, ISCSI_INITIATOR_CHAP_ID_MASK );
	HERMON_FETCH_ISCSI_AND_UPDATE( settings, &init_chapsec_setting, conf->chap_pass,
			conf->valid_mask, ISCSI_INITIATOR_CHAP_PASS_MASK );
}

static int hermon_nv_write_iscsi_config ( struct hermon *hermon, struct hermon_port *port ) {
	int rc;

	hermon_nv_write_iscsi_init_dhcp ( port );
	hermon_nv_write_iscsi_first_tgt ( port );
	hermon_nv_write_iscsi_initiator ( port );

	if ( ( rc = hermon_flash_write_iscsi_config ( hermon, port ) ) )
		DBGC ( hermon, "Failed to write iSCSI parameters to the flash (rc = %d)\n", rc );

	return rc;
}

static int
( * hermon_nv_write_settings[] ) ( struct hermon *hermon, struct hermon_port *port ) = {
	hermon_nv_write_nic_config,
	hermon_nv_write_iscsi_config,
	NULL,
};


int hermon_nv_store ( struct driver_settings *driver_settings __unused, void *priv ) {
	struct hermon *hermon = ( struct hermon * ) priv;
	struct settings *settings;
	struct hermon_nvmem_conf *conf;
	char buf[255] = {0};
	int rc, modified = 0;

	if ( ! nv_settings_root ) {
		printf ( "%s: nv_settings_root is not initialized\n", __FUNCTION__ );
		return -22;
	}

	settings = & ( nv_settings_root->settings ) ;
	conf = & ( hermon->hermon_nv_conf );

	/* Make sure flash access is supported */
        if ( ! hermon->cap.nv_mem_access_supported )
                return 0;

	/* Open hardware */
        if ( ( rc = hermon_open ( hermon ) ) != 0 )
                return rc;

	HERMON_FETCH_SETTING ( settings, &virt_mode_setting );
        if ( rc > 0 ) {
                modified = 1;
                conf->virt_conf.virt_mode = ( buf[0] == 'S' );
        }

        HERMON_FETCH_SETTING ( settings, &virt_num_setting );
        if ( rc > 0 ) {
                modified = 1;
                conf->virt_conf.num_of_vfs = strtoul ( buf, NULL, 10 );
        }

	if ( modified && ( rc = hermon_flash_write_hermon_settings ( hermon ) ) ) {
                printf ( "Failed to write general settings to the flash (rc = %d)\n", rc );
                goto err_flash_write;
        }

	rc = 0;
 err_flash_write:
	/* Close hardware */
	hermon_close ( hermon );
	return rc;
}

int hermon_nv_port_store ( struct driver_settings *settings, void *priv ) {
	struct hermon *hermon = ( struct hermon * ) priv;
	struct hermon_port *port = container_of ( settings,
			struct hermon_port, driver_settings );
	int i, rc;

	/* Make sure flash access is supported */
	if ( ! hermon->cap.nv_mem_access_supported )
		return 0;

	/* Open hardware */
	if ( ( rc = hermon_open ( hermon ) ) != 0 )
		return rc;

	for ( i = 0; hermon_nv_write_settings[i] != NULL ; i++ ) {
		if ( ( rc = hermon_nv_write_settings[i] ( hermon, port ) ) ) {
			printf ("Failed to store %s settings (sub settings - %d) in NV memory.\n",
				settings->generic_settings.settings.name, i );
		}
	}

	/* Close hardware */
	hermon_close ( hermon );

	return 0;
}

void hermon_restore_port_default ( struct driver_settings *driver_settings ) {
	struct hermon_setting_operation *op;
	struct settings *settings;
	struct extended_setting *ext;
	unsigned int i;
	int rc;

	for ( i = 0 ; i < ( sizeof ( hermon_setting_operations ) /
                            sizeof ( hermon_setting_operations[0] ) ) ; i++ ) {
		op = &hermon_setting_operations[i];
		if ( fetch_setting_origin ( &driver_settings->generic_settings.settings, op->setting, &settings ) )
			continue;
		ext = find_extended ( op->setting );
		if ( settings && ext->type != LABEL ) {
			if ( ( rc = hermon_store ( settings, op->setting, NULL, 0 ) ) )
				DBG ( "Failed to store the default value of %s (rc = %d)\n",
						op->setting->name, rc );
		}
	}
}

void hermon_restore_default ( struct driver_settings *driver_settings __unused ) {
	struct hermon_setting_operation *op;
        struct settings *settings;
        struct extended_setting *ext;
        unsigned int i;
        int rc;

	if ( ! nv_settings_root ) {
		printf ( "%s: nv_settings_root is not initialized\n", __FUNCTION__ );
		return;
	}

	settings = &nv_settings_root->settings;

	for ( i = 0 ; i < ( sizeof ( hermon_setting_operations ) /
				sizeof ( hermon_setting_operations[0] ) ) ; i++ ) {
		op = &hermon_setting_operations[i];
		if ( !setting_applies ( settings, op->setting ) )
			continue;
		ext = find_extended ( op->setting );
		if ( ext->type != LABEL ) {
			if ( ( rc = hermon_store ( settings, op->setting, NULL, 0 ) ) )
				DBG ( "Failed to store the default value of %s (rc = %d)\n",
						op->setting->name, rc );
		}
	}
}

int get_virt_id_type ( struct settings *settings ) {
	char buf[MAX_SETTING_STR] = { 0 };
	int rc;

	rc = fetchf_setting ( settings, &virt_lan_setting, NULL, NULL, buf, MAX_SETTING_STR );
	if ( ( rc > 0 ) && ( buf[0] == 'D' ) )
		return LABEL;

	return INPUT;
}

int get_virt_num_type ( struct settings *settings ) {
	char buffer[MAX_SETTING_STR] = { 0 };
	int rc;

	if ( setting_applies ( settings, &virt_mode_setting ) ) {
		rc = fetchf_setting ( settings, &virt_mode_setting, NULL, NULL, buffer, MAX_SETTING_STR );
		if ( ( rc > 0 ) && ( buffer[0] == 'S' ) )
			return INPUT;
	}

	return LABEL;
}

int hermon_init_port_settings ( struct hermon *hermon, struct hermon_port *port ) {
	struct driver_settings *driver_settings = &port->driver_settings;
	struct net_device *netdev = port->netdev;
	char *menu_name;
	char tmp[100] = { 0 };

	if ( hermon->cap.nv_mem_access_supported == 0 )
		return 0;

	sprintf ( tmp, "%s : Port %d - %s ", netdev->name, port->ibdev->port, netdev_addr( netdev ) );
	menu_name = zalloc ( ( strlen ( tmp ) + 1 ) );
	strcpy ( menu_name, tmp );

	driver_settings->priv = hermon;
	driver_settings->nv_store = &hermon_nv_port_store;
	driver_settings->set_default = &hermon_restore_port_default;

	INIT_LIST_HEAD ( &driver_settings->modified_list );
	driver_settings->store_menu = 0;
	init_driver_settings ( driver_settings, &hermon_settings_operations, menu_name );

	/* Set the function pointers of the extended settings */
	ext_virt_id.get_extended_type = &get_virt_id_type;
	ext_virt_num.get_extended_type = &get_virt_num_type;

	return 0;
}

int hermon_init_settings ( struct hermon *hermon ) {
	struct driver_settings *driver_settings = &hermon->hermon_settings;
	int rc;

	if ( hermon->cap.nv_mem_access_supported == 0 )
		return 0;

	generic_settings_init ( &hermon->hermon_settings.generic_settings, NULL);
        if ( ( rc = register_settings (&hermon->hermon_settings.generic_settings.settings,
		NULL , "System setup" ) ) != 0 )
		return rc;
        hermon->hermon_settings.generic_settings.settings.op = &hermon_settings_operations;
        hermon->hermon_settings.generic_settings.settings.default_scope = &main_scope;
	nv_settings_root = &hermon->hermon_settings.generic_settings;

	driver_settings->nv_store = &hermon_nv_store;
	driver_settings->set_default = &hermon_restore_default;
	driver_settings->priv = hermon;

	INIT_LIST_HEAD ( &driver_settings->modified_list );
	driver_settings->store_menu = 0;

	/* Add to driver settings list  */
	list_add ( &driver_settings->list, &driver_settings_list );

	init_firmware_settings ( &hermon_settings_operations );

	return 0;
}

struct hermon_device_name hermon_device_names [] = {
	{ 0x1003, "ConnectX3 HCA" },
	{ 0x1007, "ConnectX3-Pro HCA" },
};

char * get_device_name ( struct hermon *hermon ) {
	unsigned int i;
	for ( i = 0 ; i < ( sizeof ( hermon_device_names ) /
				sizeof ( hermon_device_names[0] ) ) ; i++ ) {
		if ( hermon->pci->id->device == hermon_device_names[i].device_id )
			return hermon_device_names[i].name;
	}
	return NULL;
}


